import dataclasses
import hashlib
import json
import pickle


class HashGenerator:
    """Calculate hash for the given object.

    Note that we cannot simply hashlib.sha1(pickle.dumps(obj)) since some of the contents might be cached on remote. For nested objects, we calculate hash for each child element recursively.
    """
    @classmethod
    def calculate_hash(cls, value, context=None):
        def get_hash(value):
            # import torch lazily to avoid loading torch when it is not needed.
            import torch
            from .variable import Variable
            if isinstance(value, dict):
                value = json.dumps({k: get_hash(v) for k, v in sorted(value.items())})
            elif isinstance(value, (list, tuple)):
                value = json.dumps([get_hash(v) for v in value])
            elif dataclasses.is_dataclass(value):
                value = json.dumps({k: get_hash(v) for k, v in sorted(dataclasses.asdict(value).items())})
            elif isinstance(value, Variable):
                return value.get_hash(context)
            elif isinstance(value, (str, bytes)):
                pass
            elif isinstance(value, torch.Tensor):
                value = value.cpu().detach().numpy().tobytes()
            elif hasattr(value, '__getstate__'):
                return get_hash(value.__getstate__())
            elif hasattr(value, '__reduce__'):
                try:
                    states = value.__reduce__()
                    value = json.dumps([str(value.__class__), get_hash((states[1:3]))])
                except TypeError:
                    pass

            if isinstance(value, str):
                value = value.encode('utf-8')

            if not isinstance(value, bytes):
                # Fall back to pickle.
                value = pickle.dumps(value)

            return hashlib.sha1(value).hexdigest()

        return get_hash(value)



import itertools
from .task import Task


class Job:
    def __init__(self, description):
        self._description = description
        self._tasks = [Task(t) for t in description.tasks]
        self._on_error = [Task(t) for t in description.on_error] if description.on_error else []

        # Assign uniqe names to the tasks.
        used_names = set()
        for t in itertools.chain(self._tasks, self._on_error):
            if t.name in used_names:
                count = int(t.name.split('@')[1]) if '@' in t.name else 1
                while True:
                    count += 1
                    new_name = t.name.split('@')[0] + f'@{count}'
                    if new_name not in used_names:
                        break
                t.name = new_name
            used_names.add(t.name)

    @property
    def tasks(self):
        for t in self._tasks:
            yield t

    @property
    def on_error(self):
        for t in self._on_error:
            yield t

    def load_modules(self):
        for t in itertools.chain(self.tasks, self.on_error):
            try:
                t.load_module()
            except Exception as e:
                raise RuntimeError(f"Failed to load {t.name}") from e

    def __str__(self):
        return "Job {\n" + '\n'.join([f"  {t}" for t in self.tasks]) + '\n}'



import contextvars
import dataclasses
import importlib
import importlib.util
import sys
import types


current_session = contextvars.ContextVar('current_session')


class TasksModuleStub(types.ModuleType):
    def __getattr__(self, name):
        if name.startswith('__'):
            raise AttributeError()

        stub_tasks = sys.modules['irisml.tasks']
        del sys.modules['irisml.tasks']
        original_module = importlib.import_module(f'irisml.tasks.{name}')
        sys.modules['irisml.tasks'] = stub_tasks
        return TaskStub(name, original_module.Task)


class TaskStub:
    def __init__(self, task_name, original_task):
        self._task_name = task_name
        self._task = original_task

    def __call__(self, **kwargs):
        session = current_session.get()
        name = kwargs.pop('_name', None)

        input_names = [f.name for f in dataclasses.fields(self._task.Inputs)]
        config_names = [f.name for f in dataclasses.fields(self._task.Config)]
        if invalid_names := set(kwargs.keys()) - set(input_names) - set(config_names):
            raise ValueError(f'Invalid arguments: {invalid_names}')

        inputs = {k: kwargs[k] for k in input_names if k in kwargs}
        config = {k: kwargs[k] for k in config_names if k in kwargs}
        assert len(inputs) + len(config) == len(kwargs)
        task_instance = TaskInstance()
        name = session.add_task(task_instance, self._task_name, name, inputs, config)
        return OutputsStub(task_instance, name, self._task.Outputs)

    def __str__(self):
        return f'TaskStub(task={self._task_name})'


class TaskInstance:
    def __init__(self):
        self._reference_counter = 0

    def increment_reference_counter(self):
        self._reference_counter += 1

    @property
    def ref_count(self):
        return self._reference_counter

    def __str__(self):
        return f'TaskInstance(ref_count={self._reference_counter})'


class OutputsStub:
    def __init__(self, task, name, outputs_dataclass):
        self._task = task
        self._name = name
        self._output_types = {f.name: f.type for f in dataclasses.fields(outputs_dataclass)}

    def increment_reference_counter(self):
        self._task.increment_reference_counter()

    def __getattr__(self, name):
        if name not in self._output_types:
            raise AttributeError(f'No output named {name}')
        return OutputsMemberStub(self, self._name, name, self._output_types[name])

    def __str__(self):
        return f'OutputsStub(name={self._name}, outputs={self._output_types})'


class VariableStub:
    def resolve(self):
        return str(self)


class OutputsMemberStub(VariableStub):
    def __init__(self, parent, parent_name, name, data_type):
        self._parent = parent
        self._parent_name = parent_name
        self._name = name
        self._data_type = data_type

    def resolve(self):
        self._parent.increment_reference_counter()
        return super().resolve()

    def __str__(self):
        return f'$output.{self._parent_name}.{self._name}'


class EnvStub(VariableStub):
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return f'$env.{self._name}'



import dataclasses
import importlib
import json
import pathlib
import re
import typing
import unittest
from irisml.core import TaskBase


VALID_NAME_PATTERN = re.compile('[a-z][a-z0-9_]*')


def _generate_test_task_method(filepath):
    def test_method(self):
        print(f"Testing {filepath}")
        task_name = filepath.stem
        self.assertTrue(VALID_NAME_PATTERN.match(task_name))
        task_module = importlib.import_module(f'irisml.tasks.{task_name}')
        task_class = task_module.Task

        self.assertIsNotNone(task_class)
        self.assertTrue(task_class.__doc__, "The Task class must have docstring.")
        self.assertTrue(issubclass(task_class, TaskBase))
        self.assertTrue(task_class.VERSION and task_class.VERSION != '0.0.0')
        self.assertTrue(dataclasses.is_dataclass(task_class.Config))
        self.assertTrue(dataclasses.is_dataclass(task_class.Inputs))
        self.assertTrue(dataclasses.is_dataclass(task_class.Outputs))

        config_names = [f.name for f in dataclasses.fields(task_class.Config)]
        inputs_names = [f.name for f in dataclasses.fields(task_class.Inputs)]
        outputs_names = [f.name for f in dataclasses.fields(task_class.Outputs)]

        self.assertFalse(set(config_names) & set(inputs_names))
        self.assertTrue(all(VALID_NAME_PATTERN.match(n) for n in config_names))
        self.assertTrue(all(VALID_NAME_PATTERN.match(n) for n in inputs_names))
        self.assertTrue(all(VALID_NAME_PATTERN.match(n) for n in outputs_names))
        self.assertTrue(True)

        if task_class.dry_run != TaskBase.dry_run:
            self.assertTrue(all(f.default or f.default_factory for f in dataclasses.fields(task_class.Outputs)), "If dry_run() is not defined, all Outputs fields must have default values.")

    return test_method


def _generate_test_config_method(filepath):
    def test_method(self):
        print(f"Testing {filepath}")
        original_text = filepath.read_text()
        config = json.loads(original_text)

        self.assertIsInstance(config, dict)
        self.assertIn('tasks', config)
        self.assertIsInstance(config['tasks'], list)

        self.assertFalse(set(config.keys()) - {'tasks', 'on_error'})

        # Check the format
        formatted_text = json.dumps(config, sort_keys=False, indent=4) + '\n'
        if original_text != formatted_text:
            self.fail(f"{filepath} is not formatted. Please use 'python -m json.tool' to format the file.")

    return test_method


def generate_task_tests(target_filepaths: typing.List[pathlib.Path]):
    class TestTasks(unittest.TestCase):
        pass

    for filepath in target_filepaths:
        if filepath.stem == '__pycache__':
            continue
        test_name = f'test_{filepath.stem}'
        setattr(TestTasks, test_name, _generate_test_task_method(filepath))

    return TestTasks


def generate_config_tests(target_filepaths: typing.List[pathlib.Path]):
    class TestConfigs(unittest.TestCase):
        pass

    used = set()
    for filepath in target_filepaths:
        test_name = f'test_{filepath.stem}'
        index = 2
        while test_name in used:
            test_name = f'test_{filepath.stem}_{index}'
            index += 1
        used.add(test_name)
        setattr(TestConfigs, test_name, _generate_test_config_method(filepath))

    return TestConfigs



import dataclasses
import sys
import typing
import unittest
import unittest.mock
from typing import Dict, List, Tuple, Optional
from irisml.core import Context, TaskBase, TaskDescription
from irisml.core.task import Task
from irisml.core.variable import OutputVariable


class TestTask(unittest.TestCase):
    def test_load_config(self):
        @dataclasses.dataclass
        class ChildConfig:
            int_var: int

        @dataclasses.dataclass
        class Config:
            int_var: int
            str_var: str
            optional_int_var: Optional[int]
            tuple_var: Tuple[str, str]
            int_list_var: List[int]
            str_dict_var: Dict[str, str]
            nested: ChildConfig
            optional_nested: Optional[ChildConfig]
            list_of_tuple: List[Tuple[str, float]]
            int_default_var: int = 42

        result = Task._load_config(Config, {'int_var': 12, 'str_var': '12', 'optional_int_var': 24, 'tuple_var': ('k', 'v'), 'int_list_var': [1, 2, 3], 'str_dict_var': {'k': 'v', 'k2': 'v2'},
                                            'nested': {'int_var': 123}, 'optional_nested': {'int_var': 345}, 'list_of_tuple': [('k1', 1), ('k2', 2.5)]})
        self.assertEqual(result.nested, ChildConfig(int_var=123))
        self.assertEqual(result.optional_nested, ChildConfig(int_var=345))
        self.assertEqual(result.int_var, 12)
        self.assertEqual(result.str_var, '12')
        self.assertEqual(result.optional_int_var, 24)
        self.assertEqual(result.int_list_var, [1, 2, 3])
        self.assertEqual(result.str_dict_var, {'k': 'v', 'k2': 'v2'})
        self.assertEqual(result.int_default_var, 42)
        self.assertEqual(result.tuple_var, ('k', 'v'))
        self.assertEqual(result.list_of_tuple, [('k1', 1), ('k2', 2.5)])

    def test_load_module(self):
        class FakeTask:
            class Task(TaskBase):
                pass

        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': FakeTask}):
            task = Task(TaskDescription.from_dict({'task': 'fake_task'}))
            task.load_module()

        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks': unittest.mock.MagicMock()}):
            task = Task(TaskDescription.from_dict({'task': 'fake_task'}))
            with self.assertRaisesRegex(RuntimeError, 'Task not found'):
                task.load_module()  # Since fake_task doesn't exist, it raises a RuntimeError with a custom error message.

    def test_output_variable_in_config(self):
        @dataclasses.dataclass
        class Config:
            optional_int_var: Optional[int]
            int_list_var: List[int]
            str_dict_var: Dict[str, str]
            int_default_var: int = 42

        @dataclasses.dataclass
        class Outputs:
            optional_int_var: Optional[int]
            int_list_var: List[int]
            str_dict_var: Dict[str, str]

        context = Context()
        context.add_outputs('test_output', Outputs(optional_int_var=None, int_list_var=[1, 2, 3], str_dict_var={'key': 'value'}))

        result = Task._load_config(Config, {'optional_int_var': OutputVariable('$output.test_output.optional_int_var'),
                                            'int_list_var': OutputVariable('$output.test_output.int_list_var'),
                                            'str_dict_var': OutputVariable('$output.test_output.str_dict_var')})
        resolved = context.resolve(result)
        self.assertIsNone(resolved.optional_int_var)
        self.assertEqual(resolved.int_list_var, [1, 2, 3])
        self.assertEqual(resolved.str_dict_var, {'key': 'value'})

    def test_validate(self):
        class CustomTask:
            class Task(TaskBase):
                class Outputs:  # This class is supposed to have @dataclasses.dataclass decorator.
                    int_value: int

        class EmptyClass:
            pass

        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(TaskDescription.from_dict({'task': 'custom_task'}))
            with self.assertRaises(RuntimeError):
                task.load_module()

            CustomTask.Task.Outputs = EmptyClass
            task = Task(TaskDescription.from_dict({'task': 'custom_task'}))
            with self.assertRaises(RuntimeError):
                task.load_module()

    def test_inputs_env(self):
        """Environment variable in inputs should be casted to the expected type."""
        task_description = TaskDescription.from_dict({'task': 'custom_task', 'inputs': {'input0': '$env.ENV_INPUT'}})

        class CustomTask:
            class Task(TaskBase):
                @dataclasses.dataclass
                class Inputs:
                    input0: int

                def execute(self, inputs):
                    if not isinstance(inputs.input0, int):
                        raise RuntimeError(f"Input type is {type(inputs.input0)}")
                    return self.Outputs()

        context = Context({'ENV_INPUT': '12345'})
        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(task_description)
            task.load_module()
            task.execute(context)

    def test_literal_env(self):
        # The Env variable should work for typing.Literal.
        task_description = TaskDescription.from_dict({'task': 'custom_task', 'inputs': {'input0': '$env.ENV_INPUT'}, 'config': {'config0': '$env.ENV_CONFIG', 'config1': '$env.ENV_CONFIG2'}})

        class CustomTask:
            class Task(TaskBase):
                @dataclasses.dataclass
                class Inputs:
                    input0: typing.Literal['r2d2', 'c3po']

                @dataclasses.dataclass
                class Config:
                    config0: typing.Literal['365', '42']
                    config1: typing.Literal[365, 42]

                def execute(self, inputs):
                    if inputs.input0 not in ['r2d2', 'c3po']:
                        raise RuntimeError
                    if self.config.config0 not in ['365', '42']:
                        raise RuntimeError
                    if self.config.config1 not in [365, 42]:
                        raise RuntimeError
                    return self.Outputs()

        context = Context({'ENV_INPUT': 'r2d2', 'ENV_CONFIG': '42', 'ENV_CONFIG2': '42'})
        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(task_description)
            task.load_module()
            task.execute(context)

        # Those values are not in the Literal types.
        context = Context({'ENV_INPUT': 'unknown', 'ENV_CONFIG': '111', 'ENV_CONFIG2': '111'})
        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            with self.assertRaises(ValueError):
                task = Task(task_description)
                task.load_module()
                task.execute(context)

    def test_hash_not_calculated(self):
        context = Context()  # No cache manager

        class CustomTask:
            class Task(TaskBase):
                def execute(self, inputs):
                    return self.Outputs()

        with unittest.mock.patch.dict(sys.modules), unittest.mock.patch('irisml.core.task.HashGenerator') as m_hash_generator:
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(TaskDescription.from_dict({'task': 'custom_task'}))
            task.load_module()
            task.execute(context)

            m_hash_generator.calculate_hash.assert_not_called()



import logging
import typing
from irisml.core import JobDescription
from irisml.core.cache_manager import create_storage_manager, CacheManager
from irisml.core.context import Context
from irisml.core.job import Job

logger = logging.getLogger(__name__)


class JobRunner:
    """Helper class to run a job."""
    def __init__(self, job_dict: typing.Dict, env_vars: typing.Dict[str, str], cache_storage_url: str = None, no_cache_read: bool = False):
        job_description = JobDescription.from_dict(job_dict)
        self._job = Job(job_description)
        self._env_vars = env_vars
        self._cache_storage_url = cache_storage_url
        self._no_cache_read = no_cache_read

    def run(self, dry_run=False):
        logger.debug("Loading task modules.")
        self._job.load_modules()

        logger.info("Running a job.")

        cache_manager = CacheManager(create_storage_manager(self._cache_storage_url), self._no_cache_read) if self._cache_storage_url and not dry_run else None
        if cache_manager:
            logger.info(f"Cache is enabled: {self._cache_storage_url}")

        context = Context(self._env_vars, cache_manager)

        # Note that the random seed will be reset in each Task.execute().
        for task in self._job.tasks:
            logger.debug(f"Running a task: {task}")
            try:
                task.execute(context, dry_run)
            except Exception as e:
                logger.exception(f"Failed to run a task {task}: {e}")
                self.on_error(context, e)
                raise

        logger.info("Completed.")

    def on_error(self, context, e):
        context.add_environment_variable('IRISML_EXCEPTION', e)
        for task in self._job.on_error:
            logger.info(f"Running an error handler: {task}")
            task.execute(context)



import unittest
import typing
from unittest.mock import MagicMock
from irisml.core.context import Context
from irisml.core.variable import EnvironmentVariable, OutputVariable, replace_variables


class TestEnvironmentVariable(unittest.TestCase):
    def test_invalid_names(self):
        invalid_names = ['RANDOM_NAME', '$invalid_variable_type.RANDOM_NAME', '$env.NESTED.NAME', '$env.lower_case']
        for name in invalid_names:
            with self.assertRaises(ValueError):
                EnvironmentVariable(name)

    def test_resolve(self):
        context = Context({'CORRECT_VAR': '42'})
        v = EnvironmentVariable('$env.CORRECT_VAR')
        self.assertEqual(v.resolve(context), '42')
        v.expected_type = int
        self.assertEqual(v.resolve(context), 42)

        v = EnvironmentVariable('$env.VAR_NOT_FOUND')
        with self.assertRaises(ValueError):
            v.resolve(context)

    def test_union_type(self):
        context = Context({'CORRECT_VAR': 1})
        v = EnvironmentVariable('$env.CORRECT_VAR')
        v.expected_type = typing.Union[int, str]
        self.assertEqual(type(v.resolve(context)), int)

    def test_union_type_fails(self):
        context = Context({'CORRECT_VAR': "fail"})
        v = EnvironmentVariable('$env.CORRECT_VAR')
        v.expected_type = typing.Union[int, float]
        self.assertEqual(type(v.resolve(context)), str)

    def test_non_str(self):
        context = Context({'CORRECT_VAR': RuntimeError()})
        v = EnvironmentVariable('$env.CORRECT_VAR')
        v.expected_type = Exception
        self.assertEqual(type(v.resolve(context)), RuntimeError)


class TestOutputVariable(unittest.TestCase):
    def test_invalid_names(self):
        invalid_names = ['random_name', '$invalid_variable_type.random_name', '$output.task_name_only', '$output.task_name.UPPER_CASE']
        for name in invalid_names:
            with self.assertRaises(ValueError):
                OutputVariable(name)

    def test_resolve(self):
        context = Context()
        context.add_outputs('task_name', MagicMock(int_output=42, str_output='42'))
        v = OutputVariable('$output.task_name.int_output')
        self.assertEqual(v.resolve(context), 42)
        v = OutputVariable('$output.task_name.str_output')
        self.assertEqual(v.resolve(context), '42')

        with self.assertRaises(ValueError):
            EnvironmentVariable('$output.task_name.no_output').resolve(context)
        with self.assertRaises(ValueError):
            EnvironmentVariable('$output.no_task_name.str_output').resolve(context)


class TestReplaceVariables(unittest.TestCase):
    def test_non_variable(self):
        data = {'abc': 123, 'cde': [1, 2, 3], 'key': 'value', 'empty': ''}
        self.assertEqual(replace_variables(data), data)

    def test_replace_str(self):
        self.assertIsInstance(replace_variables('$env.ENV_VAR'), EnvironmentVariable)
        self.assertIsInstance(replace_variables('$output.task_name.var'), OutputVariable)

    def test_containers(self):
        self.assertEqual(replace_variables(['$env.ENV_VAR', 42]), [EnvironmentVariable('$env.ENV_VAR'), 42])
        self.assertEqual(replace_variables({'name': '$env.ENV_VAR', 'name2': 42}), {'name': EnvironmentVariable('$env.ENV_VAR'), 'name2': 42})

    def test_invalid_name(self):
        with self.assertRaises(ValueError):
            replace_variables('$new_var')



import pathlib
import tempfile
import textwrap
import unittest
from irisml.core import JobDescription
from irisml.compiler import Compiler


class TestCompiler(unittest.TestCase):
    def test_simple(self):
        input_str = """
        from irisml.tasks import task_a, task_b

        def main():
            output = task_a()
            output = task_b(in_int=output.out_int)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [
            {'task': 'task_a'},
            {'task': 'task_b', 'inputs': {'in_int': '$output.task_a.out_int'}}]})

        self.check_compile(input_str, expected_job_description)

    def test_unique_names(self):
        input_str = """
        from irisml.tasks import task_a, task_b

        def main():
            output = task_a()
            task_a()
            output2 = task_a()
            task_b(in_int=output.out_int)
            task_b(in_int=output2.out_int)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [
            {'task': 'task_a'},
            {'task': 'task_a'},  # Doesn't have an explicit name because the output is not used.
            {'task': 'task_a', 'name': 'task_a_2'},
            {'task': 'task_b', 'inputs': {'in_int': '$output.task_a.out_int'}},
            {'task': 'task_b', 'inputs': {'in_int': '$output.task_a_2.out_int'}}]})
        self.check_compile(input_str, expected_job_description)

    def test_manual_names(self):
        input_str = """
        from irisml.tasks import task_a

        def main():
            task_a(_name='my_dataset')
        """
        expected_job_description = JobDescription.from_dict({'tasks': [{'task': 'task_a', 'name': 'my_dataset'}]})
        self.check_compile(input_str, expected_job_description)

    def test_env_variable(self):
        input_str = """
        from irisml.compiler import get_env
        from irisml.tasks import task_a

        def main():
            int_value = get_env('INT_VALUE')
            task_a(in_int=int_value)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [{'task': 'task_a', 'config': {'in_int': '$env.INT_VALUE'}}]})
        self.check_compile(input_str, expected_job_description)

    def test_function_call(self):
        inputs_str = """
        from irisml.tasks import task_a, task_b

        def my_func(int_value):
            return task_b(in_int=int_value)

        def main():
            output = task_a()
            my_func(int_value=output.out_int)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [
            {'task': 'task_a'},
            {'task': 'task_b', 'inputs': {'in_int': '$output.task_a.out_int'}}]})
        self.check_compile(inputs_str, expected_job_description)

    def test_on_error(self):
        input_str = """
        from irisml.tasks import task_a
        from irisml.compiler import on_error

        def main():
            task_a()
            on_error(lambda: task_a())
        """
        expected_job_description = JobDescription.from_dict({'tasks': [{'task': 'task_a'}], 'on_error': [{'task': 'task_a'}]})
        self.check_compile(input_str, expected_job_description)

    def test_child_tasks(self):
        input_str = """
        from irisml.compiler import make_tasks
        from irisml.tasks import task_a, task_c

        def main():
            tasks = make_tasks(lambda: task_a())
            task_c(tasks=tasks)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [{'task': 'task_c', 'config': {'tasks': [{'task': 'task_a'}]}}]})
        self.check_compile(input_str, expected_job_description)

    def test_child_tasks_with_args(self):
        input_str = """
        from irisml.compiler import make_tasks
        from irisml.tasks import task_a, task_b, task_c

        def child_tasks(int_value):
            task_b(in_int=int_value)

        def main():
            output = task_a()
            tasks = make_tasks(child_tasks, int_value=output.out_int)
            task_c(tasks=tasks)
        """
        expected_job_description = JobDescription.from_dict({'tasks': [
            {'task': 'task_a'},
            {'task': 'task_c', 'config': {'tasks': [{'task': 'task_b', 'inputs': {'in_int': '$output.task_a.out_int'}}]}}]})
        self.check_compile(input_str, expected_job_description)

    def check_compile(self, input_str, expected_job_description):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            self._make_fake_tasks(temp_dir)
            input_filepath = temp_dir / 'input.py'
            input_filepath.write_text(textwrap.dedent(input_str))
            compiler = Compiler()
            job_description = compiler.compile(input_filepath, include_paths=[])
            print(job_description)
            self.assertEqual(job_description, expected_job_description)

    def _make_fake_tasks(self, dir_path):
        tasks_dir = dir_path / 'irisml' / 'tasks'
        tasks_dir.mkdir(parents=True)

        (tasks_dir / 'task_a.py').write_text(textwrap.dedent("""
        import dataclasses
        import irisml.core

        class Task(irisml.core.TaskBase):
            @dataclasses.dataclass
            class Config:
                in_int: int = 100

            @dataclasses.dataclass
            class Outputs:
                out_int: int = 42
        """))

        (tasks_dir / 'task_b.py').write_text(textwrap.dedent("""
        import dataclasses
        import irisml.core

        class Task(irisml.core.TaskBase):
            @dataclasses.dataclass
            class Inputs:
                in_int: int
        """))

        (tasks_dir / 'task_c.py').write_text(textwrap.dedent("""
        import dataclasses
        import typing
        from irisml.core import TaskDescription
        import irisml.core

        class Task(irisml.core.TaskBase):
            @dataclasses.dataclass
            class Config:
                tasks: typing.List[TaskDescription]
        """))



import dataclasses
from typing import Any, Dict, List, Optional
import typing


def to_dict(x):
    if hasattr(x, 'to_dict'):
        return x.to_dict()
    if isinstance(x, dict):
        return {k: to_dict(v) for k, v in x.items()}
    if isinstance(x, list):
        return [to_dict(v) for v in x]
    return x


@dataclasses.dataclass
class TaskDescription:
    task: str
    name: Optional[str] = None  # Optional name for the task. Must be unique in the job. This name will be used for OutputVariable names.
    inputs: Optional[Dict[str, Any]] = None
    config: Optional[Dict[str, Any]] = None

    @classmethod
    def from_dict(cls, data: typing.Dict):
        return cls(**data)

    def to_dict(self) -> typing.Dict:
        d = {'task': self.task}
        if self.name:
            d['name'] = self.name
        if self.inputs:
            d['inputs'] = to_dict(self.inputs)
        if self.config:
            d['config'] = to_dict(self.config)
        return d


@dataclasses.dataclass
class JobDescription:
    tasks: List[TaskDescription]
    on_error: Optional[List[TaskDescription]]

    @classmethod
    def from_dict(cls, data: typing.Dict):
        c = {'tasks': [TaskDescription.from_dict(t) for t in data['tasks']],
             'on_error': data.get('on_error', None) and [TaskDescription.from_dict(t) for t in data['on_error']]}
        return cls(**c)

    def to_dict(self) -> typing.Dict:
        d = {'tasks': [t.to_dict() for t in self.tasks]}
        if self.on_error:
            d['on_error'] = [t.to_dict() for t in self.on_error]
        return d



import argparse
import json
import os
import pathlib
from irisml.core.job_runner import JobRunner
from irisml.core.commands.common import configure_logger


def main():
    class KeyValuePairAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            k, v = values.split('=', 1)
            d = getattr(namespace, self.dest)
            d[k] = v

    parser = argparse.ArgumentParser(description="Run a job")
    parser.add_argument('job_filepath', type=pathlib.Path)
    parser.add_argument('--env', '-e', default={}, action=KeyValuePairAction)
    parser.add_argument('--dry_run', '-n', action='store_true')
    parser.add_argument('--verbose', '-v', action='store_true')
    parser.add_argument('--very_verbose', '-vv', action='store_true')
    parser.add_argument('--no_cache', action='store_true')
    parser.add_argument('--no_cache_read', action='store_true', help="Don't read from cache, only write")

    args = parser.parse_args()

    configure_logger(2 if args.very_verbose else (1 if args.verbose else 0))

    cache_storage_url = (not args.no_cache) and os.getenv('IRISML_CACHE_URL')
    job_description = json.loads(args.job_filepath.read_text())
    job_runner = JobRunner(job_description, args.env, cache_storage_url=cache_storage_url, no_cache_read=args.no_cache_read)
    job_runner.run(dry_run=args.dry_run)


if __name__ == '__main__':
    main()



import copy
import importlib
import importlib.abc
import importlib.util
import json
import logging
import os
import pathlib
import subprocess
import sys
import tempfile

from irisml.core import JobDescription
import irisml.compiler.stubs


logger = logging.getLogger(__name__)


class Compiler:
    def compile(self, input_filepath, include_paths):
        include_paths = [str(pathlib.Path(p).resolve()) for p in include_paths]
        env = copy.deepcopy(os.environ)
        env['PYTHONPATH'] = ':'.join(include_paths) + ':' + env.get('PYTHONPATH', '')

        with tempfile.NamedTemporaryFile() as f:
            p = subprocess.run([sys.executable, '-c',
                                f'import irisml.compiler; irisml.compiler.compiler.init(); from {input_filepath.stem} import main; irisml.compiler.compiler.generate(main, "{f.name}")'],
                               capture_output=True, cwd=input_filepath.parent, env=env)

            json_str = pathlib.Path(f.name).read_text()

        stderr_str = p.stderr.decode()
        stdout_str = p.stdout.decode()

        if stdout_str:
            print(stdout_str)

        for s in stderr_str.splitlines():
            logger.warning(s)

        if p.returncode != 0:
            raise RuntimeError("Failed to compile.")

        try:
            job_description_dict = json.loads(json_str)
        except json.JSONDecodeError:
            raise RuntimeError(f"Failed to parse JSON: {stdout_str}")

        return JobDescription.from_dict(job_description_dict)


class CompilerSession:
    def __init__(self):
        self._tasks = []
        self._names = set()
        self._on_error = None

    def get_unique_name(self, task_name, name):
        """Get a unique name for the task."""
        if name:
            if name in self._names:
                raise ValueError(f'Name {name} is already used.')
            self._names.add(name)
            return name

        if task_name not in self._names:
            self._names.add(task_name)
            return task_name

        i = 1
        while True:
            name = f'{task_name}_{i}'
            if name not in self._names:
                self._names.add(name)
                return name
            i += 1

    def add_task(self, task_instance, task_name, name, inputs, config):
        implicit_name = self.get_unique_name(task_name, name)
        task = {'task': task_name, 'explicit_name': name, 'implicit_name': implicit_name, 'inputs': self._resolve_variable(inputs), 'config': self._resolve_variable(config),
                'task_instance': task_instance}
        self._tasks.append(task)
        return implicit_name

    def set_on_error(self, tasks):
        self._on_error = [t.to_dict() for t in tasks]

    def generate(self):
        def process_task(task):
            t = {'task': task['task'], 'inputs': task['inputs'], 'config': task['config']}
            if task['explicit_name']:
                t['name'] = task['explicit_name']
            elif task['task_instance'].ref_count > 0 and task['implicit_name'] != task['task']:
                t['name'] = task['implicit_name']
            return t

        return JobDescription.from_dict({'tasks': [process_task(t) for t in self._tasks], 'on_error': self._on_error})

    @staticmethod
    def _resolve_variable(data):
        if isinstance(data, irisml.compiler.stubs.VariableStub):
            return data.resolve()
        if isinstance(data, dict):
            return {k: CompilerSession._resolve_variable(v) for k, v in data.items()}
        if isinstance(data, list):
            return [CompilerSession._resolve_variable(v) for v in data]
        if isinstance(data, irisml.compiler.stubs.OutputsStub):
            raise TypeError(f"OutputsStub is not allowed in the inputs/config: {data}")
        return data


class StubImporter(importlib.abc.MetaPathFinder, importlib.abc.Loader):
    def __init__(self):
        self._enabled = True

    def find_spec(self, fullname, path, target=None):
        if fullname == 'irisml.tasks' and self._enabled:
            # Replace the tasks module with a stub only once.
            self._enabled = False
            self._original_tasks = None
            return importlib.util.spec_from_loader(fullname, self)
        return None

    def create_module(self, spec):
        return irisml.compiler.stubs.TasksModuleStub(spec.name)

    def exec_module(self, module):
        pass  # Do nothing


def init():
    """Initialize the environment for compiler."""
    sys.meta_path.insert(0, StubImporter())


def generate(main, output_filename):
    """Generate a job description from a main function."""
    session = CompilerSession()
    irisml.compiler.stubs.current_session.set(session)
    main()
    irisml.compiler.stubs.current_session.set(None)

    job_description_dict = session.generate().to_dict()

    pathlib.Path(output_filename).write_text(json.dumps(job_description_dict, indent=4))



import unittest
from irisml.core import JobDescription
from irisml.core.job import Job


class TestJob(unittest.TestCase):
    def test_assign_unique_name(self):
        job_description = {'tasks': [
            {'task': 'custom_task'},
            {'task': 'custom_task'},
            {'task': 'custom_task'},
            {'task': 'custom_task', 'name': 'custom_name'},
            {'task': 'custom_task', 'name': 'custom_name'},
            {'task': 'custom_task', 'name': 'custom_name'}
        ]}

        job = Job(JobDescription.from_dict(job_description))
        names = [t.name for t in job.tasks]
        self.assertEqual(names, ['custom_task', 'custom_task@2', 'custom_task@3', 'custom_name', 'custom_name@2', 'custom_name@3'])



import typing

from irisml.compiler.compiler import CompilerSession
from irisml.compiler.stubs import EnvStub, current_session


def get_env(name: str):
    return EnvStub(name)


def make_tasks(func: typing.Callable, **kwargs):
    old_session = current_session.get()
    session = CompilerSession()
    current_session.set(session)
    func(**kwargs)
    current_session.set(old_session)

    job_description = session.generate()
    return [t.to_dict() for t in job_description.tasks]


def on_error(func: typing.Callable):
    old_session = current_session.get()
    session = CompilerSession()
    current_session.set(session)
    func()
    current_session.set(old_session)

    job_description = session.generate()
    if not job_description.tasks:
        raise RuntimeError("on_error must have at least one task")

    old_session.set_on_error(job_description.tasks)



import argparse
import json
import pathlib
from irisml.compiler import Compiler


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_filepath', type=pathlib.Path)
    parser.add_argument('--output_filepath', '-o', type=pathlib.Path, required=True)
    parser.add_argument('-I', action='append', dest='include_paths', default=[])

    args = parser.parse_args()

    compiler = Compiler()
    job_description = compiler.compile(args.input_filepath, args.include_paths)
    args.output_filepath.parent.mkdir(parents=True, exist_ok=True)
    args.output_filepath.write_text(json.dumps(job_description.to_dict(), indent=4) + '\n')


if __name__ == '__main__':
    main()



import argparse
import datetime
import os
import pathlib
import pickle
import re
from azure.identity import DefaultAzureCredential
from azure.storage.blob import ContainerClient


def cache_list(blobs):
    for b in blobs:
        print(f"{b.name}: last_modified_at: {b.last_modified}")

    print(f"Total: {len(blobs)} blobs.")


def cache_remove(blobs, container_client):
    response = input(f"Removing {len(blobs)} blobs. Continue? [y/N]: ")
    if response.lower() != 'y':
        return

    chunk_size = 256
    for i in range(0, len(blobs), chunk_size):
        container_client.delete_blobs(*blobs[i:i+chunk_size])
    print("Deleted.")


def cache_show(blobs, container_client):
    for b in blobs:
        if b.size < 1024:
            data = pickle.loads(container_client.download_blob(b).readall())
            print(f"{b.name}: {data}")
        else:
            print(f"{b.name}: size={b.size}. To download the file, please use the download command.")


def cache_download(blobs, container_client):
    response = input(f"Found {len(blobs)} blobs. Download all? [y/N]: ")
    if response.lower() != 'y':
        return

    for b in blobs:
        local_filepath = pathlib.Path(b.name)
        local_filepath.parent.mkdir(parents=True, exist_ok=True)
        local_filepath.write_bytes(container_client.download_blob(b).readall())
        print(f"Saved {local_filepath}")


def get_blobs(container_client, mtime, name):
    if mtime:
        m = re.match(r'([+-])([0-9]+)$', mtime)
        assert m.group(1) in ('+', '-')
        time_threshold = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=int(m.group(2)))
        time_older = m.group(1) == '+'
    else:
        time_threshold = None

    def match_condition(b):
        if time_threshold:
            is_older = b.last_modified < time_threshold
            if time_older != is_older:  # XOR
                return False

        if name and name not in b.name:
            return False

        return True

    return list(b for b in container_client.list_blobs() if match_condition(b))


def main():
    parser = argparse.ArgumentParser(description="Manage a cache storage.")
    subparser = parser.add_subparsers(dest='command', required=True)

    def add_subparser(subparser, name):
        s = subparser.add_parser(name)
        s.add_argument('--mtime', help="Select blobs that were modified n*24 hours ago.")
        s.add_argument('--name', help="Select blobs whose name contain this string.")

    add_subparser(subparser, 'download')
    add_subparser(subparser, 'list')
    add_subparser(subparser, 'remove')
    add_subparser(subparser, 'show')

    args = parser.parse_args()

    cache_storage_url = os.getenv('IRISML_CACHE_URL')
    if not cache_storage_url:
        parser.error("Please set IRISML_CACHE_URL.")

    print(f"Using cache url: {cache_storage_url}")

    container_client = ContainerClient.from_container_url(cache_storage_url, credential=DefaultAzureCredential())
    blobs = get_blobs(container_client, args.mtime, args.name)
    if not blobs:
        print("No blob found.")
        return

    if args.command == 'list':
        cache_list(blobs)
    elif args.command == 'download':
        cache_list(blobs)
        cache_download(blobs, container_client)
    elif args.command == 'remove':
        cache_list(blobs)
        cache_remove(blobs, container_client)
    elif args.command == 'show':
        cache_show(blobs, container_client)


if __name__ == '__main__':
    main()
